import time
import hashlib
from flask_sqlalchemy import SQLAlchemy
from authlib.integrations.sqla_oauth2 import (
    OAuth2ClientMixin,
    OAuth2AuthorizationCodeMixin,
    OAuth2TokenMixin,
)

db = SQLAlchemy()


class User(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    username = db.Column(db.String(40), unique=True, nullable=False)
    password = db.Column(db.BLOB, unique=False, nullable=False)
    salt = db.Column(db.BLOB, unique=False, nullable=False)

    name = db.Column(db.String(40), unique=False)
    surname = db.Column(db.String(40), unique=False)

    scope = db.Column(db.String(40), unique=False, nullable=False)
    uuid = db.Column(db.String(50), unique=True, nullable=False)

    def __str__(self):
        return self.username

    def get_user_id(self):
        return self.id

    def check_password(self, password):
        hashed = hashlib.pbkdf2_hmac(
            'sha256', password.encode('utf-8'), self.salt, 100000)
        return self.password == hashed


class AccessCode(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    code = db.Column(db.String(50), unique=True, nullable=False)
    scope = db.Column(db.String(40), unique=False, nullable=False)
    user_id = db.Column(
        db.Integer, db.ForeignKey('user.id'))
    user = db.relationship('User')


class ColorAllocation(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    day = db.Column(db.Date, unique=True, nullable=False)
    color_id = db.Column(
        db.Integer, db.ForeignKey('color.id', ondelete='CASCADE'))
    color = db.relationship('Color')


class Color(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    hex = db.Column(db.String(50), nullable=False)
    inverse = db.Column(db.Boolean, nullable=False)
    name = db.Column(db.String(50), unique=False)


class OAuth2Client(db.Model, OAuth2ClientMixin):
    __tablename__ = 'oauth2_client'

    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(
        db.Integer, db.ForeignKey('user.id', ondelete='CASCADE'))
    user = db.relationship('User')


class OAuth2AuthorizationCode(db.Model, OAuth2AuthorizationCodeMixin):
    __tablename__ = 'oauth2_code'

    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(
        db.Integer, db.ForeignKey('user.id', ondelete='CASCADE'))
    user = db.relationship('User')


class OAuth2Token(db.Model, OAuth2TokenMixin):
    __tablename__ = 'oauth2_token'

    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(
        db.Integer, db.ForeignKey('user.id', ondelete='CASCADE'))
    user = db.relationship('User')

    def is_refresh_token_active(self):
        if self.revoked:
            return False
        expires_at = self.issued_at + self.expires_in * 2
        return expires_at >= time.time()
