import 'dart:convert';

import 'package:http/http.dart' as http;

import '../models/access_code.dart';
import '../models/color_allocation.dart';
import '../models/ribbon_color.dart';
import '../models/user.dart';

class AdminApi {
  final String _baseUrl;
  final http.Client _client;

  AdminApi(this._client, this._baseUrl);

  http.Client get client => _client;

  Future<AccessCode?> addAccessCode(String code, String scope) async {
    final response = await _client.post(
      Uri.parse('$_baseUrl/access_codes'),
      body: json.encode({'code': code, 'scope': scope}),
      headers: {'Content-Type': 'application/json'},
    );

    if (response.statusCode == 200) {
      return AccessCode.fromMap(json.decode(response.body));
    } else {
      return null;
    }
  }

  Future<RibbonColor?> addColor(RibbonColor color) async {
    final response = await _client.post(
      Uri.parse('$_baseUrl/colors'),
      body: json.encode({
        'hex': color.hex.toRadixString(16),
        if (color.name != null) 'name': color.name,
        'inverse': color.inverse,
      }),
      headers: {'Content-Type': 'application/json'},
    );

    if (response.statusCode == 200) {
      return RibbonColor.fromMap(json.decode(response.body));
    } else {
      return null;
    }
  }

  Future<ColorAllocation?> addColorAllocation(ColorAllocation alloc) async {
    final response = await _client.post(
      Uri.parse('$_baseUrl/color_allocations'),
      body: json.encode(alloc.toMap),
      headers: {'Content-Type': 'application/json'},
    );

    if (response.statusCode == 200) {
      return ColorAllocation.fromMap(json.decode(response.body));
    } else {
      return null;
    }
  }

  Future<User?> addUser(User user, String password) async {
    final body = user.toMap;
    body.addAll({'password': password});
    final response = await _client.post(
      Uri.parse('$_baseUrl/users'),
      body: json.encode(body),
      headers: {'Content-Type': 'application/json'},
    );

    if (response.statusCode == 200) {
      return User.fromMap(json.decode(response.body));
    } else {
      return null;
    }
  }

  Future<void> deleteAccessCode(int codeId) async {
    final response = await _client.delete(Uri.parse('$_baseUrl/access_codes/$codeId'));

    if (response.statusCode == 200) {
      return;
    }
  }

  Future<void> deleteColor(int colorId) async {
    final response = await _client.delete(Uri.parse('$_baseUrl/colors/$colorId'));

    if (response.statusCode == 200) {
      return;
    }
  }

  Future<void> deleteColorAllocation(int colorAllocationId) async {
    final response =
        await _client.delete(Uri.parse('$_baseUrl/color_allocations/$colorAllocationId'));

    if (response.statusCode == 200) {
      return;
    }
  }

  Future<void> deleteUser(int userId) async {
    final response = await _client.delete(Uri.parse('$_baseUrl/users/$userId'));

    if (response.statusCode == 200) {
      return;
    }
  }

  Future<List<AccessCode>?> getAccessCodes() async {
    final response = await _client.get(Uri.parse('$_baseUrl/access_codes'));

    if (response.statusCode == 200) {
      return (json.decode(response.body) as List).map((e) => AccessCode.fromMap(e)).toList();
    } else {
      return null;
    }
  }

  Future<List<ColorAllocation>?> getColorAllocations() async {
    final response = await _client.get(Uri.parse('$_baseUrl/color_allocations'));

    if (response.statusCode == 200) {
      return (json.decode(response.body) as List).map((e) => ColorAllocation.fromMap(e)).toList();
    } else {
      return null;
    }
  }

  Future<List<RibbonColor>?> getColors() async {
    final response = await _client.get(Uri.parse('$_baseUrl/colors'));

    if (response.statusCode == 200) {
      return (json.decode(response.body) as List).map((e) => RibbonColor.fromMap(e)).toList();
    } else {
      return null;
    }
  }

  Future<ColorAllocation?> getCurrentColor() async {
    final response = await _client.get(Uri.parse('$_baseUrl/current_color_admin'));

    if (response.statusCode == 200) {
      return ColorAllocation.fromMap(json.decode(response.body));
    }
  }

  Future<User?> getUser(int userId) async {
    final response = await _client.get(Uri.parse('$_baseUrl/users/$userId'));

    if (response.statusCode == 200) {
      return User.fromMap(json.decode(response.body));
    } else {
      return null;
    }
  }

  Future<List<User>?> getUsers() async {
    final response = await _client.get(Uri.parse('$_baseUrl/users'));

    if (response.statusCode == 200) {
      return (json.decode(response.body) as List).map((e) => User.fromMap(e)).toList();
    } else {
      return null;
    }
  }

  Future<RibbonColor?> updateColor(RibbonColor color) async {
    final response = await _client.put(
      Uri.parse('$_baseUrl/colors/${color.id}'),
      body: json.encode(color.toMap),
      headers: {'Content-Type': 'application/json'},
    );

    if (response.statusCode == 200) {
      return RibbonColor.fromMap(json.decode(response.body));
    } else {
      return null;
    }
  }

  Future<ColorAllocation?> updateColorAllocation(ColorAllocation alloc) async {
    final response = await _client.put(
      Uri.parse('$_baseUrl/color_allocations/${alloc.id}'),
      body: json.encode(alloc.toMap),
      headers: {'Content-Type': 'application/json'},
    );

    if (response.statusCode == 200) {
      return ColorAllocation.fromMap(json.decode(response.body));
    } else {
      return null;
    }
  }

  Future<User?> updateUser(User user) async {
    final response = await _client.put(
      Uri.parse('$_baseUrl/users/${user.id}'),
      body: json.encode(user.toMap),
      headers: {'Content-Type': 'application/json'},
    );

    if (response.statusCode == 200) {
      return User.fromMap(json.decode(response.body));
    } else {
      return null;
    }
  }
}
