import 'package:digital_3g_common/backend_switcher.dart';
import 'package:digital_3g_common/theme.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:flutter_localizations/flutter_localizations.dart';
import 'package:flutter_phoenix/flutter_phoenix.dart';
import 'package:http/http.dart';
import 'package:intl/intl.dart';
import 'package:logging/logging.dart';
import 'package:provider/provider.dart';
import 'package:shared_preferences/shared_preferences.dart';

import 'api/auth_api.dart';
import 'config.dart';
import 'pages/access_codes_page.dart';
import 'pages/color_allocations_page.dart';
import 'pages/colors_page.dart';
import 'pages/home_page.dart';
import 'pages/login_page.dart';
import 'pages/register_page.dart';
import 'pages/settings_page.dart';
import 'pages/users_page.dart';

void main() {
  if (debug) {
    Logger.root.level = Level.FINER;
    Logger.root.onRecord.listen((record) {
      // ignore: avoid_print
      print(
        '[${record.loggerName}] (${record.level.name}) ${record.time.toString().split('.')[0]}: ${record.message}',
      );
    });
  }

  runApp(Phoenix(child: const MyApp()));
}

class MyApp extends StatefulWidget {
  const MyApp({Key? key}) : super(key: key);

  @override
  State<MyApp> createState() => _MyAppState();
}

class _MyAppState extends State<MyApp> {
  BackendSwitcher? _bloc;

  @override
  Widget build(BuildContext context) {
    Intl.defaultLocale = 'de';
    return FutureBuilder(
      future: SharedPreferences.getInstance(),
      builder: (BuildContext context, AsyncSnapshot<SharedPreferences> data) {
        if (data.hasData) {
          final mode = data.data!.getString('digital_3g_backend_mode');
          _bloc = BackendSwitcher(
            production: baseUrl,
            development: debugBaseUrl,
            mode: stringToMode(mode ?? 'production'),
          );
          return BlocProvider<BackendSwitcher>(
            create: (BuildContext context) => _bloc!,
            child: BlocBuilder<BackendSwitcher, String>(
              bloc: _bloc,
              builder: (BuildContext context, String state) => Provider(
                create: (_) => AuthApi(Client(), state),
                child: MaterialApp(
                  title: '3-(di)G(i) Management',
                  localizationsDelegates: const [
                    GlobalMaterialLocalizations.delegate,
                    GlobalCupertinoLocalizations.delegate,
                  ],
                  supportedLocales: [Locale(Intl.defaultLocale!)],
                  debugShowCheckedModeBanner: false,
                  theme: themeData,
                  initialRoute: '/',
                  routes: {
                    '/': (BuildContext context) => const HomePage(isAdmin: false),
                    '/admin': (BuildContext context) => const HomePage(isAdmin: true),
                    '/login': (BuildContext context) => const LoginPage(),
                    '/register': (BuildContext context) => const RegisterPage(),
                    '/admin/users': (BuildContext context) => const UsersPage(),
                    '/admin/colors': (BuildContext context) => const ColorsPage(),
                    '/admin/color_allocations': (BuildContext context) =>
                        const ColorAllocationsPage(),
                    '/admin/users/codes': (BuildContext context) => const AccessCodesPage(),
                    '/settings': (BuildContext context) => const SettingsPage(),
                  },
                ),
              ),
            ),
          );
        }
        return Container();
      },
    );
  }
}
