import 'package:digital_3g_common/backend_switcher.dart';
import 'package:digital_3g_common/ui.dart';
import 'package:flutter/material.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:provider/provider.dart';

import '../api/admin_api.dart';
import '../api/auth_api.dart';
import '../models/ribbon_color.dart';
import '../widgets/card_list_tile.dart';
import '../widgets/color_dialog.dart';
import '../widgets/delete_dialog.dart';

class ColorsPage extends StatefulWidget {
  const ColorsPage({Key? key}) : super(key: key);

  @override
  _ColorsPageState createState() => _ColorsPageState();
}

class _ColorsPageState extends State<ColorsPage> {
  AdminApi? _adminApi;

  List<RibbonColor> _colors = [];

  @override
  Widget build(BuildContext context) {
    return Scaffold(
      appBar: AppBar(
        title: const Text('Farben'),
        actions: [IconButton(onPressed: _onEditColor, icon: const Icon(Icons.add))],
      ),
      body: Center(
        child: ConstrainedBox(
          constraints: const BoxConstraints(maxWidth: 640.0),
          child: RefreshIndicator(
            onRefresh: _onUpdateColors,
            child: ListView(
              padding: const EdgeInsets.all(8.0),
              children: _colors
                  .map<Widget>(
                    (RibbonColor e) => CardListTile(
                      leading: ColorCircle(
                        color: Color(e.hex),
                        inverse: e.inverse,
                        radius: 24.0,
                        padding: const EdgeInsets.all(8.0),
                      ),
                      onLongPress: () => showDialog(
                        context: context,
                        builder: (BuildContext context) => SimpleDialog(
                          children: [
                            ListTile(
                              title: const Text('Invertiert duplizieren'),
                              onTap: () async {
                                Navigator.of(context).pop();
                                await _onDuplicateInverted(e);
                              },
                            ),
                            ListTile(
                              title: const Text('Löschen'),
                              onTap: () async {
                                Navigator.of(context).pop();
                                await _onDeleteColor(e);
                              },
                            ),
                          ],
                        ),
                      ),
                      title: Text(e.toRadixString),
                      subtitle: Text(
                        e.inverse ? 'Invertiertes Logo' : 'Normales Logo',
                        style: Theme.of(context).textTheme.subtitle2,
                      ),
                      actions: [
                        IconButton(
                          onPressed: () => _onEditColor(color: e),
                          icon: const Icon(Icons.edit),
                        ),
                      ],
                    ),
                  )
                  .toList(),
            ),
          ),
        ),
      ),
    );
  }

  @override
  void initState() {
    super.initState();
    _initApi();
  }

  void _initApi() async {
    final _client = Provider.of<AuthApi>(context, listen: false).client;
    _adminApi = AdminApi(_client, BlocProvider.of<BackendSwitcher>(context).state);

    _onUpdateColors();
    super.didChangeDependencies();
  }

  Future<void> _onDeleteColor(RibbonColor e) async {
    final result = await showDialog(
      context: context,
      builder: (BuildContext context) => const DeleteDialog(),
    );

    if (result) {
      await _adminApi!.deleteColor(e.id!);
      setState(() {
        _colors.removeWhere((RibbonColor c) => c.id == e.id);
      });
    }
  }

  Future<void> _onDuplicateInverted(RibbonColor color) async {
    if (_colors.where((e) => e.hex == color.hex && e.inverse == !color.inverse).isNotEmpty) {
      ScaffoldMessenger.of(context).showSnackBar(
        const SnackBar(content: Text('Diese Farbe existiert bereits in invertierter Form')),
      );
      return;
    }
    final newColor = await _adminApi!.addColor(color.copyWith(inverse: !color.inverse));
    if (newColor != null) {
      setState(() => _colors.add(newColor));
    } else {
      ScaffoldMessenger.of(context).showSnackBar(
        const SnackBar(content: Text('Die Farbe konnte nicht dupliziert werden')),
      );
    }
  }

  void _onEditColor({RibbonColor? color}) async {
    final result = await showDialog<ColorDialogResult>(
      context: context,
      builder: (BuildContext context) => ColorDialog(color: color),
    );

    if (!(result?.save ?? false)) return;

    if (color != null) {
      setState(() {
        color = result!.color;
      });
      await _adminApi!.updateColor(color!);
    } else {
      _colors.add((await _adminApi!.addColor(result!.color))!);
      setState(() {
        _colors;
      });
    }
  }

  Future<void> _onUpdateColors() async {
    _colors = (await _adminApi!.getColors()) ?? [];
    _colors.sort((a, b) => a.hex.compareTo(b.hex));
    setState(() => _colors);
  }
}
